/* Copyright (c) 1994, 1995, 1996, Takayuki Hosoda. All rights reserved. */

/* #define */
#include "feal.h"
#include "fnxcrypt.h"

char           *errors[] =
{
    "Process complete.\n",
    "encrypt/decrypt a file using FEAL-NX Version 1.3.9 (Sep. 11 1996)\n\
Copyright (c) 1994, 1995, 1996, Takayuki Hosoda. All rights reserved.\n\
fnxcrypt is free software with ABSOLUTELY NO WARRANTY.\n\
Synopsis :\n\
  fnxcrypt -h\n\
  fnxcrypt [-e|-d] [-N] [-cbc|-cfb] [-f] [readfile [writefile]] [-kkeyfile]\n\
Options :\n\
  -N   : iteration value which should be one of 8, 16, 32(default) or 64.\n\
  -cbc : use cipher block chaining mode. (default)\n\
  -cfb : use cipher feed back mode.\n\
  -d   : decrypt with password.\n\
  -e   : encrypt with password.\n\
  -f   : attempt to overwrite existing writefile.\n\
  -k   : use keyfile for password.\n\
  -h   : show this message.\n",
    "Too few or too many arguments.\n",
    "Incorrect option.\n",
    "Cannot open write file.\n",
    "Cannot open read file.\n",
    "File write error.\n",
    "Not enough memory.\n",
    "Unconditional end of file.\n",
    "Output file already exists.\n",
    "Process aborted.\n",
    "Internal error.\n"
};

/* errmsg */
int
errmsg(err)
    int             err;
{
    fprintf(stderr, "%s%s", "fnxcrypt: ", errors[err]);
    return (err);
}

#ifdef BSD
long
get_time()
{
    struct timeval  tv;

    gettimeofday(&tv, NULL);
    return (tv.tv_sec);
}
#else
long
get_time()
{
    time_t         *tp = NULL;

    return (time(tp));
}
#endif

/* make_iv */
UCHAR          *
make_iv(iv)
    UCHAR          *iv;
{
    long            l;
    int             i;
    static UCHAR    randnum[16] =
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    l = get_time();
    for (i = 0; i < 4; i++) {
	randnum[i] ^= (unsigned char) (l & 0x0ff);
	l >>= 8;
    }
    if (fealskey(randnum, 64))
	exit(errmsg(E_INTERN));
    fealnx(iv, randnum, 64);
    fealnx(randnum, iv, 64);
    return (iv);
}
/* get_pass */
char *
get_pass()
{
    UCHAR           verify[16];
    char           *password = NULL;
    char           *cp;
    int             len = 0;
    int             retype = S_GET;

    mem_set(verify, 0, 16);
    while (retype) {
	switch (retype) {
	case S_RETYPE:
	    cp = password = getpass("Retype password:");
	    retype = ((*password == '\0') || (password == NULL)) ? S_ABORT :
		((mem_cmp(verify, (UCHAR *) password, 16) ? S_MISMAT : S_GOT));
	    break;
	case S_GET:
	    mem_set(verify, 0, 16);
	    cp = password = getpass("Enter password:");
	    for (len = 0; ((len < 16) && (*cp++)); len++);
	    mem_cpy(verify, (UCHAR *) password, len);
	    retype = ((*password == '\0') || (password == NULL)) ? S_ABORT :
		((len < 6) ? S_SHORT : S_RETYPE);
	    break;
	case S_SHORT:
	    fprintf(stderr, "Please enter a longer password.\n");
	    retype = S_GET;
	    break;
	case S_MISMAT:
	    fprintf(stderr, "Mismatch; try again, Enter to quit.\n");
	    retype = S_GET;
	    break;
	case S_ABORT:
	    return (NULL);
	default:
	    break;
	}
    }
    cp = NULL;
    mem_set(verify, 0, 16);
    return (password);
}

/* fnxcrypt */
USHORT
fnxcrypt(fdr, fdw, fdk, mode, n)
    int             fdr;
    int             fdw;
    int             fdk;
    int             mode;
    int             n;		/* FEAL iteration value */
{
    /* variables */
    UCHAR           in[8];
    UCHAR           out[8];
    UCHAR           fb[8];
    UCHAR           key[16];
    register int    i;
    int             r;
    char           *password;
    char           *cp;
    short           rest;

    mem_set(fb, 0, 8);

    /* randomize initial value. */
    if (mode & ENCRYPT)
	make_iv(fb);

    /* get password */
    mem_set(key, 0, 16);
    if (mode & KEY_FILE) {
	if(fdk == -1)
	    exit(errmsg(E_INTERN));
	xread(fdk, key, 16);
	close(fdk);
    } else {
	if (NULL == (cp = password = get_pass()))
	    return (E_ABORT);
	for (i = 0; ((i < 16) && (*cp++)); i++);
	mem_cpy(key, password, i);
        mem_set((UCHAR *) password, 0, i);	/* clear password */
    }
    /* randomize initial value again. */
    if (mode & ENCRYPT)
	make_iv(fb);

    /* make expanded key */
    if (fealskey(key, n))
	exit(errmsg(E_INTERN));
    mem_set(key, 0, 16);	/* clear key */

    /* process data */
    if (mode & ENCRYPT) {
	/* write initial value */
	if(NO_ERR != (r = wfifo(fb, 8, fdw)))
	    return(r);
	if (mode & CBC) {
	    /* encrypt in CBC mode */
	    while (8 == (rest = rfifo(in, 8, fdr))) {
		for (i = 0; i < 8; i++)
		    in[i] ^= fb[i];
		fealnx(in, fb, n);
		if(NO_ERR != (r = wfifo(fb, 8, fdw)))
		    return(r);
	    }
	    if (rest) {
		for (i = (int) rest; i < 8; i++)
		    in[i] = 0;
		for (i = 0; i < 8; i++)
		    in[i] ^= fb[i];
		fealnx(in, fb, n);
		if(NO_ERR != (r = wfifo(fb, 8, fdw)))
		    return(r);
		fb[0] = (UCHAR) (rest & 0x07);
		if(NO_ERR != (r = wfifo(fb, 1, fdw)))
		    return(r);
	    }
	    /* end CBC */
	} else {
	    /* encrypt in CFB mode */
	    while (1 == rfifo(in, 1, fdr)) {
		fealnx(fb, out, n);
		out[0] ^= in[0];
		for (i = 0; i < 7; i++)
		    fb[i] = fb[i + 1];
		fb[7] = out[0];
		if(NO_ERR != (r = wfifo(out, 1, fdw)))
		    return(r);
	    }
	    /* end CFB */
	}
    } else {
	/* read initial value */
	if (8 != rfifo(fb, 8, fdr))
	    return (E_EOF);
	if (mode & CBC) {
	    /* decrypt in CBC mode */
	    if (8 != rfifo(in, 8, fdr))
		return (E_EOF);
	    fealnx(in, out, -n);
	    for (i = 0; i < 8; i++) {
		out[i] ^= fb[i];
		fb[i] = in[i];
	    }
	    while (8 == (rest = rfifo(in, 8, fdr))) {
		if(NO_ERR != (r = wfifo(out, 8, fdw)))
		    return(r);
		fealnx(in, out, -n);
		for (i = 0; i < 8; i++) {
		    out[i] ^= fb[i];
		    fb[i] = in[i];
		}
	    }
	    if (rest) {
		if(NO_ERR != (r = wfifo(out, (USHORT) in[0], fdw)))
		    return(r);
	    } else {
		if(NO_ERR != (r = wfifo(out, 8, fdw)))
		    return(r);
	    }
	    /* end CBC */
	} else {
	    /* decrypt in CFB mode */
	    while (1 == rfifo(in, 1, fdr)) {
		fealnx(fb, out, n);
		out[0] ^= in[0];
		for (i = 0; i < 7; i++)
		    fb[i] = fb[i + 1];
		fb[7] = in[0];
		if(NO_ERR != (r = wfifo(out, 1, fdw)))
		    return(r);
	    }
	    /* end CFB */
	}
    }

    /* clear expanded key */
    if (fealskey(key, n))
	exit(errmsg(E_INTERN));
    if(NO_ERR != (r = wfifo(out, FLUSHBUF, fdw)))
        return(r);
    return(NO_ERR);
}

/* parse */
int
parse(argc, argv, rd_file, wr_file, key_file, filec, mode, n)
    int             argc;
    char           *argv[];
    char          **rd_file;
    char          **wr_file;
    char          **key_file;
    int            *filec;
    int            *mode;
    int            *n;
{
    int             i;
    int             rtnv;
    int             f_flag;

    rtnv = NO_ERR;
    *filec = 0;
    *rd_file = *wr_file = *key_file = NULL;
    *mode = ENCRYPT | CBC;
    *n = 32;			/* FEAL iteration value */
    f_flag = FLAG_CLR;

    if (argc == 1) {
	rtnv = H_USAGE;
    } else {
	for (i = 1; i < argc; i++) {
	    if (*argv[i] == SWITCHAR) {
		switch (*++argv[i]) {
		case 'h':
		    rtnv = H_USAGE;
		    break;
		case 'd':
		    *mode &= ~ENCRYPT;
		    break;
		case 'e':
		    *mode |= ENCRYPT;
		    break;
		case 'f':
		    *mode |= FORCE;
		    break;
		case 'c':
		    if (*++argv[i] == 'f')
			*mode &= ~CBC;
		    break;
		case 'k':
		    *key_file = ++argv[i];
		    *mode |= KEY_FILE;
		    break;
		case '8':
		    *n = 8;
		    break;
		case '1':
		    *n = 16;
		    break;
		case '3':
		    *n = 32;
		    break;
		case '6':
		    *n = 64;
		    break;
		default:
		    rtnv = E_OPTION;
		    break;
		}
		--argv[i];
	    } else {
		if (f_flag == FLAG_CLR) {
		    f_flag |= FLAG_RDF;
		    *rd_file = argv[i];
		    *filec = 1;
		} else {
		    *wr_file = argv[i];
		    *filec += 1;
		}
	    }
	}
    }
    if (*filec >= 3)
	rtnv = E_ARG;
    return (rtnv);
}

/* main */
int
main(argc, argv)
    int             argc;
    char           *argv[];

{
    int             r;
    char           *rd_file;
    char           *wr_file;
    char           *key_file;
    int             filec;
    int             fdr;
    int             fdw;
    int             fdk;
    int             mode;
    int             n;

    r = parse(argc, argv, &rd_file, &wr_file, &key_file, &filec, &mode, &n);
    if (r == H_USAGE)
	exit(errmsg(H_USAGE));
    if (!((r == E_ARG) || (r == E_OPTION))) {
	if (filec >= 1) {
	    if (-1 == (fdr = open(rd_file, O_RDONLY | O_BINARY, PERMS)))
		exit(errmsg(E_FOPENR));
	} else {
	    fdr = STDIN;
	}
	fdw = (filec == 2 ? xopenw(wr_file, mode) : STDOUT);
	fdk = -1;
	if (NULL != key_file)
	    if (-1 == (fdk = open(key_file, O_RDONLY | O_BINARY, PERMS)))
		exit(errmsg(E_FOPENR));
	r = fnxcrypt(fdr, fdw, fdk, mode, n);
	close(fdw);
	close(fdr);
    }
    exit(r == NO_ERR ? r : errmsg(r));
}
