/* Copyright (c) 1994, 1995, 1996, Takayuki Hosoda. All rights reserved. */
/* fnxcrypt version 1.3.9 */

/* This program is for BSD 4.3 UNIX or MS-DOS and it might not be portable. */

#include <stdio.h>
#include "feal.h"

#ifdef MSDOS
#include <time.h>
#include <sys\timeb.h>
#else
#include <sys/times.h>
#endif

#define  ITERATION 102400

extern unsigned short k[];

int
mem_cmp(d, s, n)
    register unsigned char *d;
    register unsigned char *s;
    register int    n;
{
    for (; n > 0; n--)
	if (*d++ != *s++)
	    return (1);
    return (0);
}

int
result(tv, out, cp)
    unsigned char  *tv;
    unsigned char  *out;
    char           *cp;
{
	
	int				err;
    static char     s[17];
    static char     d[17];
   
    if(err = mem_cmp(tv, out, 8)){
    sprintf(s, "%02x%02x%02x%02x%02x%02x%02x%02x",
	    tv[0], tv[1], tv[2], tv[3], tv[4], tv[5], tv[6], tv[7]);
    sprintf(d, "%02x%02x%02x%02x%02x%02x%02x%02x",
	    out[0], out[1], out[2], out[3], out[4], out[5], out[6], out[7]);
    fprintf(stdout, "%s expected, %s found. %s FAIL.\n", s, d, cp);
    }
    return (err);
}

int
test()
{
    int             n;
    long            i;
    int             err;
    static unsigned char key[16];
    static unsigned char p[8];
    static unsigned char q[8];
    static unsigned char c[8];

    static unsigned char tstvct[7][8] = {
    	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
    	{0xce, 0xef, 0x2c, 0x86, 0xf2, 0x49, 0x07, 0x52},
    	{0x92, 0xbe, 0xb6, 0x5d, 0x0e, 0x93, 0x82, 0xfb},
    	{0x01, 0xa9, 0x43, 0x83, 0xeb, 0x19, 0xba, 0x07},
    	{0x9c, 0x9b, 0x54, 0x97, 0x3d, 0xf6, 0x85, 0xf8},
    	{0xe2, 0xb0, 0xf1, 0xc2, 0x98, 0xeb, 0x50, 0x30},
		{0x07, 0x0b, 0xba, 0xf1, 0xb4, 0x96, 0xc7, 0xba}
	};
    err = 0;
    for (i = 0; i < 8; i++)
	p[i] = q[i] = 0;
    for (i = 0; i < 16; i++)
	key[i] = 0;
    key[0] = 0x01;
    key[1] = 0x23;
    key[2] = 0x45;
    key[3] = 0x67;
    key[4] = 0x89;
    key[5] = 0xab;
    key[6] = 0xcd;
    key[7] = 0xef;

    n = 8;
    fealskey(key, n);
    fealnx(tstvct[0], c, n);
    err |= result(tstvct[1], c, "feal-8 encryption");
    fealnx(tstvct[1], p, -n);
    err |= result(tstvct[0], q, "feal-8 decryption");

    for (i = 0; i < 8; i++)
	key[i + 8] = key[i];

    n = 8;
    fealskey(key, n);
    fealnx(tstvct[0], c, n);
    err |= result(tstvct[2], c, "feal-8x encryption");
    fealnx(tstvct[2], p, -n);
    err |= result(tstvct[0], p, "feal-8x decryption");

    n = 16;
    fealskey(key, n);
    fealnx(tstvct[0], c, n);
    err |= result(tstvct[3], c, "feal-16x encryption");
    fealnx(tstvct[3], p, -n);
    err |= result(tstvct[0], p, "feal-16x decryption");

    n = 32;
    fealskey(key, n);
    fealnx(tstvct[0], c, n);
    err |= result(tstvct[4], c, "feal-32x encryption");
    fealnx(tstvct[4], p, -n);
    err |= result(tstvct[0], p, "feal-32x decryption");

    n = 64;
    fealskey(key, n);
    fealnx(tstvct[0], c, n);
    err |= result(tstvct[5], c, "feal-64x encryption");
    fealnx(tstvct[5], p, -n);
    err |= result(tstvct[0], p, "feal-64x decryption");

    fealnx(tstvct[0], c, n);
    fealnx(c, q, n);
    fealnx(q, c, n);
    fealnx(c, q, n);
    fealnx(q, c, n);
    err |= result(tstvct[6], c, "5-feal-64x encryption");
    fealnx(c, p, -n);
    fealnx(p, q, -n);
    fealnx(q, p, -n);
    fealnx(p, q, -n);
    fealnx(q, p, -n);
    err |= result(tstvct[0], p, "5-feal-64x decryption");
    if(!err)
        fprintf(stderr, "testfeal: All test vector passed.\n");
    return (err);
}

#ifdef MSDOS
int
benchmrk()
{
    int             n;
    long            i;
    unsigned char   key[16];
    struct timeb    before;
    struct timeb    after;
    unsigned char   p[8];
    unsigned char   c[8];
    short           scale;

#ifdef LSI_C
    /* It's a hack. ftime() of LSI_C has an incompatibility. */
    /* It returns 10 milli-second value for millitm balue.   */
    scale = 10;
#else
    scale = 1;
#endif

    fprintf(stderr, "testfeal: Now doing speed measurement.\n");
    for (n = 8; n <= 64; n <<= 1) {
	printf("FEAL-%dX %ld [blocks]: ", n, ITERATION);

        printf("setkey:");
	ftime(&after);
	do{ftime(&before);}while(before.millitm == after.millitm);
	ftime(&before);
	for (i = 0; i < ITERATION; i++)
	    fealskey(key, n);
	ftime(&after);
	printf("%ld ", (after.time - before.time) * 1000L
	+ (long) (scale * ((short) after.millitm - (short) before.millitm)));

	printf("encrypt:");
	do{ftime(&before);}while(before.millitm == after.millitm);
	ftime(&before);
	for (i = 0; i < ITERATION; i++)
	    fealnx(p, c, n);
	ftime(&after);
	printf("%ld ", (after.time - before.time) * 1000L
	+ (long) (scale * ((short) after.millitm - (short) before.millitm)));

	printf("decrypt:");
	do{ftime(&before);}while(before.millitm == after.millitm);
	ftime(&before);
	for (i = 0; i < ITERATION; i++)
	    fealnx(p, c, -n);
	ftime(&after);
	printf("%ld ", (after.time - before.time) * 1000L
	+ (long) (scale * ((short) after.millitm - (short) before.millitm)));
	printf("[msec]\n");
    }
}
#else
int
benchmrk()
{
    int             n;
    long            i;
    unsigned char   key[16];
    struct tms      before;
    struct tms      after;
    unsigned char   p[8];
    unsigned char   c[8];

    fprintf(stderr, "testfeal: Now doing speed measurement.\n");
    for (n = 8; n <= 64; n <<= 1) {
	printf("FEAL-%dX %ld [blocks]: ", n, ITERATION);

	printf("setkey:");
	times(&before);
	for (i = 0; i < ITERATION; i++)
	    fealskey(key, n);
	times(&after);
	printf("%ld ", after.tms_utime - before.tms_utime);

	printf("encrypt:");
	times(&before);
	for (i = 0; i < ITERATION; i++)
	    fealnx(p, c, n);
	times(&after);
	printf("%ld ", after.tms_utime - before.tms_utime);

	printf("decrypt:");
	times(&before);
	for (i = 0; i < ITERATION; i++)
	    fealnx(p, c, -n);
	times(&after);
	printf("%ld ", after.tms_utime - before.tms_utime);
	printf("[uticks]\n");
    }
}
#endif

int
main()
{
    int             err;

    err = test();
    if (!err)
	benchmrk();
    return (err);
}
